import os
from typing import List

import torch.multiprocessing as mp
from toolz import partition_all
from tqdm import tqdm
from transformers import MarianMTModel, MarianTokenizer

from datawords.constants import ROMANCE_SUPPORT


class Translator:
    ROMANCE_SUPPORT = ROMANCE_SUPPORT

    def __init__(
        self,
        source,
        target,
        *,
        model_path=None,
        limit_text=250,
        max_length=512,
        progress_bar=True,
    ):
        """
        :param source: lang source, it could be any ROMANCE lang, or `en`.
        :type source: str
        :param target: Any ROMANCE or en lang.
        :type target: str

        """

        if model_path:
            fullpath = model_path
        else:
            fullpath = build_model_name(source, target)
        self._source = source
        self._target = target
        self._limit = limit_text
        self.tokenizer = MarianTokenizer.from_pretrained(
            fullpath, model_max_length=max_length, truncation=True
        )
        self.model = MarianMTModel.from_pretrained(fullpath)
        self._progress_bar = not progress_bar

    def flag_text_from_english(self, text: str):
        to = f">>{self._target}<< {text}"
        return to

    def _translate(self, src_text: str):
        """spanish/fr,etc to EN"""
        translated = self.model.generate(
            **self.tokenizer(src_text, return_tensors="pt", padding=True)
        )
        rsp = [self.tokenizer.decode(t, skip_special_tokens=True) for t in translated]
        return rsp, translated

    def transform(self, X: List[str]) -> List[str]:
        translated = []
        for txt in tqdm(X, disable=self._progress_bar):
            final_txt = txt
            if self._source == "en":
                final_txt = self.flag_text_from_english(txt)
            s, _ = self._translate(final_txt[: self._limit])
            translated.append(s[0])
        return translated

    def transform_with_tensors(self, X: List[str]):
        translated = []
        tensors = []
        for (
            ix,
            txt,
        ) in tqdm(enumerate(X)):
            final_txt = txt
            if self._source == "en":
                final_txt = self.flag_text_from_english(txt)
            s, t = self._translate(final_txt[: self._limit])
            translated.append(s[0])
            tensors.append(t[0])
        return translated, tensors


def _transform_wrapper(model, chunk, return_list):
    for ix_text in chunk:
        rsp = model.transform([ix_text[1]])
        return_list.append((ix_text[0], rsp[0]))


def transform_mp(source, target, *, texts, n_jobs=2, model_path=None):
    """
    It uses the multiprocessing module of pytorch to work
    in multiple process
    """
    os.environ["OMP_NUM_THREADS"] = "1"
    os.environ["MKL_NUM_THREADS1"] = "1"
    mp.set_start_method("spawn", force=True)
    manager = mp.Manager()
    return_list = manager.list()
    # fixes bug
    if len(texts) < n_jobs:
        n_jobs = len(texts)

    tns = Translator(source, target, model_path=model_path)
    procs = []
    X = [(ix, x) for ix, x in enumerate(texts)]
    _chunk_size = int(len(X) / n_jobs)
    chunks = partition_all(_chunk_size, X)
    for chunk in chunks:
        proc = mp.Process(
            target=_transform_wrapper,
            args=(
                tns,
                chunk,
                return_list,
            ),
        )
        proc.start()
        procs.append(proc)

    for p in procs:
        p.join()

    reordered_texts = sorted(return_list, key=lambda tup: tup[0])

    return [t[1] for t in reordered_texts]


def build_model_name(source, target):
    _src = source
    _target = target
    if source in ROMANCE_SUPPORT:
        _src = "ROMANCE"
    elif target in ROMANCE_SUPPORT:
        _target = "ROMANCE"
    else:
        raise NameError(f"Model not found for {source} nor {target}")
    # return f"{base_path}/{model_name}"
    return f"Helsinki-NLP/opus-mt-{_src}-{_target}"



ALPHANUMERIC_REGEX = r"[a-zA-Z0-9:/\#]+"
# ALPHANUMERIC_ACCENT_REGEX = r"[A-zÀ-ú0-9:/\-\_\#]+"
ALPHANUMERIC_ACCENT_REGEX = r"[A-zÀ-ú0-9]+"
WORDS_ACCENT_REGEX = r"[A-zÀ-ú:/\#-]+"
WORDS_REGEX = r"[a-zA-Z:/\#]+"
URL_REGEX = r"https?://\S+"

CONNECTOR_WORDS = {
    "en": frozenset(
        {
            "a",
            "an",
            "and",
            "at",
            "by",
            "for",
            "from",
            "in",
            "of",
            "on",
            "or",
            "the",
            "to",
            "with",
            "without",
        }
    ),
    "es": frozenset(
        {
            "un",
            "una",
            "y",
            "para",
            "por",
            "en",
            "el",
            "asi",
            "con",
            "si",
            "asi",
            "aunque",
            "de",
            "sino",
            "pero",
            "sin",
            "como",
            "segun",
            "ahora",
            "que",
            "cuando",
            "durante",
            "entonces",
            "hasta",
            "luego",
        }
    ),
}
ROMANCE_SUPPORT = set(
    [
        "fr",
        "fr_BE",
        "fr_CA",
        "fr_FR",
        "wa",
        "frp",
        "oc",
        "ca",
        "rm",
        "lld",
        "fur",
        "lij",
        "lmo",
        "es",
        "es_AR",
        "es_CL",
        "es_CO",
        "es_CR",
        "es_DO",
        "es_EC",
        "es_ES",
        "es_GT",
        "es_HN",
        "es_MX",
        "es_NI",
        "es_PA",
        "es_PE",
        "es_PR",
        "es_SV",
        "es_UY",
        "es_VE",
        "pt",
        "pt_br",
        "pt_BR",
        "pt_PT",
        "gl",
        "lad",
        "an",
        "mwl",
        "it",
        "it_IT",
        "co",
        "nap",
        "scn",
        "vec",
        "sc",
        "ro",
        "la",
    ]
)



import re
import warnings
from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Optional, Set

import emoji
import unidecode
from pydantic import BaseModel

from datawords import _utils, constants


class ParserProto(ABC):
    """ Abstract class that any parser should agree with. """
    @abstractmethod
    def parse(self, txt: str) -> List[str]:
        pass


class Entity(BaseModel):
    txt: str
    start_char: int
    end_char: int
    label: str


class ParserConf(BaseModel):
    """
    Related to :meth:`doc_parser`
    """
    lang: str = "en"
    emo_codes: bool = False
    strip_accents: bool = False
    lower: bool = True
    numbers: bool = True
    parse_urls: bool = False
    stopw_path: Optional[str] = None
    use_stemmer: bool = False
    phrases_model_path: Optional[str] = None


def load_stop(base_path, lang="en", strip_accents=True) -> Set[str]:
    """
    Open a list of stop words.
    The final path will be:
         models/stop_es.txt
    """
    warnings.warn("load_stop() will be deprecated", DeprecationWarning)
    with open(f"{base_path}/stop_{lang}.txt", "r") as f:
        stopwords = f.readlines()

    if strip_accents:
        stop = {unidecode.unidecode(s.strip()) for s in stopwords}
    else:
        stop = {s.strip() for s in stopwords}

    # logger.debug("%d stop words loaded for lang %s", len(stop), lang)

    return stop


def load_stop2(lang="en", *, models_path=None, strip_accents=True) -> Set[str]:
    """
    Open a list of stop words.

    If `models_path` is ommited then it will look internally
    for the list of words. Actually, it supports **en**, **pt** and **es**
    
    """
    if models_path:
        fp = f"{models_path}/stop_{lang}.txt"
    else:
        fp = f"{_utils.pkg_route()}/files/stop_{lang}.txt"

    with open(fp, "r") as f:
        stopwords = f.readlines()

    if strip_accents:
        stop = {unidecode.unidecode(s.strip()) for s in stopwords}
    else:
        stop = {s.strip() for s in stopwords}

    # logger.debug("%d stop words loaded for lang %s", len(stop), lang)

    return stop


def apply_regex(reg_expr, word) -> str:
    _found = re.findall(reg_expr, word)
    if len(_found) > 0:
        if len(_found) > 1:
            word = "".join(_found)
        else:
            word = _found[0]
        return word
    return ""


def norm_token(tkn: str) -> str:
    """
    An opinated token normalizer. It lower any string, strip any accents
    and keeps letters and number from a token.
    """
    final = unidecode.unidecode(tkn.lower().strip())
    # final = apply_regex(r"[\w]", tkn)
    final = apply_regex(constants.ALPHANUMERIC_REGEX, tkn)
    return final


def doc_parser(
    txt: str,
    stop_words: Set[str],
    stemmer: Optional[Any] = None,
    emo_codes=False,
    strip_accents=True,
    lower=True,
    numbers=False,
    parse_urls=False,
) -> List[str]:
    """
    Get a string text an return a list of words.

    .. note::
        emo_codes and parse_urls alter the order of the tokens.
        If they are found, then it will put them at the end of the list.


    This function is related to :class:`ParserConf`

    :param txt: text to be parsed.
    :type txt: str
    :param stop_words: a list of stop words.
        It's possible to get a list using :meth:`load_stop2`
    :type stop_words: Set[str]
    :param stemmer: optional, a stemmer.
    :type stemmer: Optional[Any]
    :param emo_codes: if true, emo_codes will be decoded into text
    :type emo_codes: bool
    :param stip_accents: replace accents with the same letter without accent
    :type strip_accents: bool
    :param lower: transform text to lower letters.
    :type lower: bool
    :param numbers: if True it will keep numbers.
    :type numbers: bool
    :param parse_urls: keep urls.
    :type parse_urls: bool
    :return: a list of tokens
    :rtype: List[str]

    """
    # from nltk.stem import SnowballStemmer
    # text = re.sub(r"<br>+", "", txt)
    # text = txt
    text = re.sub(constants.URL_REGEX, "", txt, flags=re.MULTILINE)

    if strip_accents:
        text = unidecode.unidecode(text)
    if lower:
        text = text.lower()

    _doc = []

    for tkn in text.split():
        # ("<[^>]*>", "")
        word = tkn
        norm = norm_token(tkn)
        if norm and norm not in stop_words:
            if numbers:
                regex_parser = constants.ALPHANUMERIC_ACCENT_REGEX
            else:
                regex_parser = constants.WORDS_ACCENT_REGEX
            word = apply_regex(regex_parser, word)
            if word:
                if stemmer:
                    word = stemmer.stem(word)
                _doc.append(word)

    if emo_codes:
        _codes = emoji.demojize(text)
        _emojis = re.findall(r"\:[a-zA-Z_-]+\:", _codes)
        _doc.extend(_emojis)

    if parse_urls:
        urls = re.findall(constants.URL_REGEX, txt)
        _doc.extend(urls)

    return _doc


# def generate_ngrams(s: str, n: int = 2, numbers=True, lower=True, sep=" ") -> List[str]:
def generate_ngrams(tokens: List[str], n: int = 2, sep=" ") -> List[str]:
    """
    Generate n grams from a string `s`.

    :param tokens: a list of words already parsed.
    :type tokens: List[str]
    :param n: how many ngrams generate. 2 by default.
    :type n: int
    :param sep: Field to use as seperator between words.
    :type sep: str
    :return: a list of ngrams
    :rtype: List[str]

    """
    # Use the zip function to help us generate n-grams
    # Concatentate the tokens into ngrams and return
    ngrams = zip(*[tokens[i:] for i in range(n)])
    return [f"{sep}".join(ngram) for ngram in ngrams]


class SentencesParser(ParserProto):
    def __init__(
        self,
        lang="en",
        lower=True,
        phrases_model=None,
        stemmer: Optional[Any] = None,
        emo_codes=False,
        strip_accents=False,
        numbers=True,
        stop_words=Set[str],
    ):
        self._lang = lang
        self._lower = lower
        self._phrases = phrases_model
        self._stemmer = stemmer
        self._emo_codes = emo_codes
        self._strip_accents = strip_accents
        self._numbers = numbers
        self._stopw = stop_words

    def parse(self, txt: str) -> List[str]:
        words = doc_parser(
            txt,
            self._stopw,
            stemmer=self._stemmer,
            emo_codes=self._emo_codes,
            strip_accents=self._strip_accents,
            lower=self._lower,
            numbers=self._numbers,
        )
        if self._phrases:
            return self._phrases[words]
        return words


class SentencesIterator:
    def __init__(self, data: Iterable, *, parser: ParserProto):
        self._parser = parser
        self._data = data

    def parser_generator(self):
        for txt in self._data:
            yield self._parser.parse(txt)

    def __iter__(self):
        self._generator = self.parser_generator()
        return self

    def __next__(self):
        result = next(self._generator)
        if not result:
            raise StopIteration
        return result



import json
import os
import sqlite3
from typing import Any, Callable, Dict, List, Union

import numpy as np
from annoy import AnnoyIndex
from pydantic import BaseModel

from datawords import _utils, parsers
from datawords.models import Word2VecHelper


class LiteDoc(BaseModel):
    """
    Used by :class:`SQLiteIndex`
    """

    id: str
    text: str


class TextIndexMeta(BaseModel):
    name: str
    words_model_path: str
    vector_size: int = 100
    ann_trees: int = 10
    distance_metric: str = "angular"
    version: str = _utils.get_version()


class TextIndex:
    def __init__(
        self,
        words_model: Word2VecHelper,
        id_mapper: Dict[int, Any],
        ann_trees: int = 10,
        distance_metric: str = "angular",
        ix: AnnoyIndex = None,
    ):
        """
        This Index is opinated, usually AnnoyIndex allows any kind of Vector,
        insted, this index is prepared to accept only texts, parse it and
        perform searchs.

        AnnoyIndex only accepts integers as indices of the data,
        for that reason and id_mapper is built in the build process.


        :param words_model: Words Model to be used for vectorizing texts.
        :type words_model: Word2VecHelper
        :param id_mapper: A dict where keys are the indices stored in Annoy
            and values are the real indices in the domain which the data
            belongs to.
        :type id_mapper: Dict[int, Any]
        :param distance_metric: type of distance to be used: "angular",
            "euclidean", "manhattan", "hamming", or "dot".
        :type distance_metric: str
        :param ix: AnnoyIndex object.
        :type ix: AnnoyIndex

        """
        self.ix: AnnoyIndex = ix
        self.id_mapper = id_mapper
        self.words = words_model
        self._ann_trees = ann_trees
        self._distance_metric = distance_metric
        self._vector_size = words_model.vector_size

    def encode(self, txt: str) -> np.ndarray:
        """a wrapper around the encode method from Word2Vec model.
        it's get a txt and return a the vectorized version of the text.

        :param txt: full string sentence.
        :type txt: str
        :return: an array from text
        :rtype: np.ndarray
        """
        return self.words.encode(txt)

    def search(self, txt: str, top_n=5, include_distances=False):
        """
        After AnnoyIndex was trained, it can performs searchs over the index.
        It will returns the original id's using the id_mapper built during the
        training process.

        :param txt: full text to search
        :type txt: str
        :param top_n: How many results it returns.
        :type top_n: int
        :param include_distances: include distances.
        """
        v = self.encode(txt)
        vectors = self.ix.get_nns_by_vector(
            v, top_n, include_distances=include_distances
        )
        if include_distances:
            _res = [self.id_mapper[_v] for _v in vectors[0]]
            res = list(zip(_res, vectors[1]))
        else:
            res = [self.id_mapper[_v] for _v in vectors]
        return res

    @classmethod
    def build(
        cls,
        ids: List[Any],
        *,
        getter: Callable,
        words_model_path: str = None,
        words_model: Word2VecHelper = None,
        distance_metric: str = "angular",
        ann_trees: int = 10,
        n_jobs: int = -1,
    ) -> "TextIndex":
        """
        Build the TextIndex . Use as follows:

        .. code-block:: python

            def getter(id_: str) -> str:
                return db.get(id_)

            ix = TextIndex.build(ids, getter=getter, words_model=wv)


        Check the test cases for a better example.


        :param ids: a list of the original ids. This id's will be mapped
            with the internal ids used by Annoy.
        :type ids: List[Any]
        :param getter: A function which get's an ID and returns a texts
            to be encoded and indexed.
        :type getter: Callable
        :param words_model_path: A fulpath to the
            :class:`datawords.words.Word2VecHelper`
        :param words_model: optionally a
            :class:`datawords.words.Word2VecHelper` could be provided.
        :type words_model: Word2VecHelper
        :param vector_size: size of the vector to be indexed, it should
            match with vector_size of the word2vec model.
        :param distance_metric: type of distance to be used: "angular",
            "euclidean", "manhattan", "hamming", or "dot".
        :type distance_metric: str
        :param ann_trees: builds a forest of n_trees trees. More trees
            gives higher precision when querying
        :type ann_trees: int
        :param n_jobs: specifies the number of threads used to build the
            trees. n_jobs=-1 uses all available CPU cores.
        :type n_jobs: int
        :return: TextIndex trained.
        :rtype: TextIndex

        """
        if not words_model:
            words_model = Word2VecHelper.load(words_model_path, keyed_vectors=True)
        ix = AnnoyIndex(words_model.vector_size, distance_metric)
        id_mapper = {}
        for _ix, _id in enumerate(ids):
            data = getter(_id)
            v = words_model.encode(data)
            ix.add_item(_ix, v)
            id_mapper[_ix] = _id
        ix.build(ann_trees, n_jobs=n_jobs)
        obj = cls(
            words_model=words_model,
            id_mapper=id_mapper,
            ann_trees=ann_trees,
            distance_metric=distance_metric,
            ix=ix,
        )
        return obj

    @classmethod
    def load(
        cls, fp: Union[str, os.PathLike], words_model: Word2VecHelper = None
    ) -> "TextIndex":
        """loads the TextIndex model.

        :param fp: The path to the index. Each model is stored in a folder.
        The path should be to that folder.
        :type fp: Union[str, os.PathLike]
        :param words_model: Optional, the words model to be used.
        :type words_model: Word2VecHelper

        :return: TextIndex loaded.
        :rtype: TextIndex
        """
        name = str(fp).rsplit("/", maxsplit=1)[1]
        with open(f"{fp}/{name}.json", "r") as f:
            jmeta = json.loads(f.read())
            meta = TextIndexMeta(**jmeta)
        ix = AnnoyIndex(meta.vector_size, meta.distance_metric)
        ix.load(f"{fp}/{name}.ann")

        with open(f"{fp}/{name}.map.json", "r") as f:
            idm = json.loads(f.read())

        id_mapper = {int(k): v for k, v in idm.items()}

        if not words_model:
            words_model = Word2VecHelper.load(meta.words_model_path)

        obj = cls(
            words_model=words_model,
            id_mapper=id_mapper,
            ann_trees=meta.ann_trees,
            distance_metric=meta.distance_metric,
            ix=ix,
        )
        return obj

    def save(self, fp: Union[str, os.PathLike]):
        """
        Save index to a folder.

        :param fp: The path to the index. Each model is stored in a folder.
            The path should be to that folder.
        :type fp: Union[str, os.PathLike]
        """
        name = str(fp).rsplit("/", maxsplit=1)[1]
        _utils.mkdir_p(fp)
        self.ix.save(f"{fp}/{name}.ann")
        conf = TextIndexMeta(
            name=name,
            words_model_path=self.words.loaded_from,
            vector_size=self._vector_size,
            ann_tress=self._ann_trees,
            distance_metric=self._distance_metric,
        )

        with open(f"{fp}/{name}.json", "w") as f:
            f.write(conf.json())

        with open(f"{fp}/{name}.map.json", "w") as f:
            f.write(json.dumps(self.id_mapper))


class SQLiteIndex:
    def __init__(self, sqlite=":memory:", stopwords=set()):
        """
        SQLiteIndex allows to store documents and search over them.
        It uses the fts5 module from sqlite. Also it parse the text
        in an opinated way.

        :param sqlite: path where database will be stored.
            By default it's saved on memory
        :type sqlite: str
        :param stopwords: list of stop words to be used by the parsed.
        :type stopwords: Set[str]
        """
        self.db = sqlite3.connect(sqlite)
        self._create_tables()
        self._stopw = stopwords

    def _create_tables(self):
        cur = self.db.cursor()
        cur.execute(
            'create virtual table search_docs using fts5(id, text, tokenize="ascii");'
        )
        cur.execute(
            """CREATE TABLE search_index
        (id INTEGER PRIMARY KEY,doc_id TEXT NOT NULL UNIQUE);"""
        )
        cur.close()

    def _parse(self, txt) -> List[str]:
        tokens = parsers.doc_parser(
            txt,
            self._stopw,
            emo_codes=False,
            strip_accents=True,
            lower=True,
            numbers=True,
            parse_urls=False,
        )
        return tokens

    def _insert(self, cur, doc: LiteDoc):
        tokens = self._parse(doc.text)
        words = " ".join(tokens)
        cur.execute(
            "insert into search_index (doc_id) values (?);",
            (doc.id,),
        )
        cur.execute(
            "insert into search_docs (id, text) values (?, ?);",
            (
                doc.id,
                words,
            ),
        )

    def add(self, doc: LiteDoc) -> bool:
        """
        Adds a document of type :class:`LiteDoc` into the index.
        If the document already exist, then it will not be stored.

        :param doc: A document
        :type doc: LiteDoc
        :return: True if it was stored or false if not.
        :rtype: bool
        """
        cur = self.db.cursor()
        added = False
        try:
            self._insert(cur, doc)

            self.db.commit()
            added = True
        except sqlite3.IntegrityError:
            pass
        except sqlite3.OperationalError:
            pass

        cur.close()
        return added

    def add_batch(self, docs: List[LiteDoc]) -> List[bool]:
        """
        Add documents in batch.

        :param docs: A list of documents
        :type docs: List[LiteDoc]
        :return: True if it was stored or false if not.
        :rtype: List[bool]
        """
        cur = self.db.cursor()
        tracking = []
        for doc in docs:
            try:
                self._insert(cur, doc)
                tracking.append(True)
            except sqlite3.IntegrityError:
                tracking.append(False)
        self.db.commit()
        cur.close()
        return tracking

    def _list(self, cur, limit=10, offset=0, table="search_index"):
        result = cur.execute(f"select * from {table} LIMIT {offset}, {limit};")
        return result

    def list_docs(self, limit=10, offset=0) -> List[LiteDoc]:
        cur = self.db.cursor()
        rows = self._list(cur, limit=limit, offset=offset, table="search_docs")
        docs = [LiteDoc(id=r[0], text=r[1]) for r in rows]
        cur.close()
        return docs

    def list_ids(self, limit=10, offset=0) -> List[str]:
        """
        It lists the ids of the document stored.
        """
        cur = self.db.cursor()
        rows = self._list(cur, limit=limit, offset=offset, table="search_index")
        ids = [r[1] for r in rows]
        cur.close()
        return ids

    @property
    def total(self) -> int:
        """Returns the total of documents stored in the index."""
        cur = self.db.cursor()
        res = cur.execute("select count(*) from search_index;").fetchone()
        cur.close()
        return res[0]

    def search(self, text: str, limit: int = 1) -> List[LiteDoc]:
        """
        Performs a search in the index.

        :param text: text to search.
        :type text: str
        :param limit: how many results retrieve.
        :type limit: int
        :return: Documents found
        :rtype: List[LiteDoc]
        """
        tokens = self._parse(text)
        words = " ".join(tokens)
        cur = self.db.cursor()
        try:
            result = cur.execute(
                f"""select * from search_docs where text MATCH '"{words}" *'
                                    limit {limit}"""
            ).fetchall()

        except sqlite3.OperationalError:
            result = []
        cur.close()
        return [LiteDoc(id=r[0], text=r[1]) for r in result]



from datawords import parsers


def open_texts():
    with open("tests/texts.txt", "r") as f:
        texts = f.readlines()
    for t in texts:
        _t = t.strip()
        if _t:
            yield _t


def test_parsers_generate_ngrams():
    ngrams = parsers.generate_ngrams(["hello", "Tom", "how"])

    assert len(ngrams) == 2


def test_parsers_doc_parser():
    stopw = parsers.load_stop2()
    t = """Goodbye world, Hi Fernández. http://chuchu.me/spotify  방 #EEER 😋.\n
        This is the 99th case for 99 days"""

    default = parsers.doc_parser(
        t, stopw, emo_codes=False, strip_accents=True, numbers=False, parse_urls=False
    )
    _all = parsers.doc_parser(
        t, stopw, emo_codes=True, strip_accents=False, numbers=True, parse_urls=True
    )

    assert len(default) == 8
    # assert "th" not in default # it fails now
    assert "this" not in default
    assert "á" not in default[3]
    assert default[0][0] == "g"
    assert "99" in _all
    assert ":face_savoring_food:" in _all
    assert "http://chuchu.me/spotify" in _all



from annoy import AnnoyIndex

from datawords import parsers
from datawords.indexes import LiteDoc, SQLiteIndex, TextIndex
from datawords.models import Word2VecHelper


def open_texts():
    with open("tests/texts.txt", "r") as f:
        texts = f.readlines()
    for t in texts:
        _t = t.strip()
        if _t:
            yield _t


def test_indexes_words_index():
    texts = open_texts()
    parser_conf = parsers.ParserConf(lang="en")
    # model = PhrasesModel(parser_conf)
    # model.fit(texts)
    elements = {x[0]: x[1] for x in enumerate(list(open_texts()))}

    def getter(id_):
        return elements[id_]

    wv = Word2VecHelper(parser_conf)
    wv.fit(texts)
    # WordsIndex(words_model=wv, id_mapper={})
    ids = list(elements.keys())
    ix = TextIndex.build(ids, getter=getter, words_model=wv)
    assert isinstance(ix.ix, AnnoyIndex)


def test_indexes_sqlite_search():
    texts = list(open_texts())
    stopw = parsers.load_stop2()
    docs = [LiteDoc(id=ix, text=t) for ix, t in enumerate(texts[:5])]
    ix = SQLiteIndex(stopwords=stopw)
    ix.add_batch(docs)
    total = ix.total
    res = ix.search("coco", limit=1)
    assert isinstance(res[0], LiteDoc)
    assert total == 5



import os
import shlex
import subprocess
import sys

TORCH_CPU_WHEEL = "https://download.pytorch.org/whl/cpu"
TORCH_CUDA11_6_WHEEL = "https://download.pytorch.org/whl/cu116"


def check_python_version():
    """
    Makes sure that the script is run with Python 3.7 or newer.
    """
    if sys.version_info.major == 3 and sys.version_info.minor >= 7:
        return
    version = "{}.{}".format(sys.version_info.major, sys.version_info.minor)
    raise RuntimeError(
        "Unsupported Python version {}. "
        "Supported versions: 3.7 - 3.11".format(version)
    )


def shell(
    command: str, check=True, input=None, cwd=None, env=None
) -> subprocess.CompletedProcess:
    """
    Runs a provided command, streaming its output to the log files.
    :param command: A command to be executed, as a single string.
    :param check: If true, will throw exception on failure (exit code != 0)
    :param input: Input for the executed command.
    :param cwd: Directory in which to execute the command.
    :param env: A set of environment variable for the process to use.
        If None, the current env is inherited.
    :return: CompletedProcess instance - the result of the command execution.
    """
    proc = subprocess.run(
        shlex.split(command),
        check=check,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        input=input,
        cwd=cwd,
        env=env,
    )

    return proc


def install_pytorch(pkg, pkg_version, cuda_version=None):
    cmd = f"pip install {pkg}=={pkg_version}"
    error = False
    if cuda_version == "11.6":
        print(f"==> Installing {pkg} ver: {pkg_version} with support for cuda 11.6")
        cmd = f"{cmd} --extra-index-url {TORCH_CUDA11_6_WHEEL}"
        msg = f"{pkg}=={pkg_version} cuda:{cuda_version} succesfully installed!!"
    elif cuda_version == "11.7":
        print(f"==> Installing {pkg} ver: {pkg_version} with support for cuda 11.7")
        msg = f"{pkg}=={pkg_version} cuda:{cuda_version} succesfully installed!!"

    elif not cuda_version:
        print(f"==> Installing {pkg} ver: {pkg_version} only for CPU")
        cmd = f"{cmd} --extra-index-url {TORCH_CPU_WHEEL}"
        msg = f"{pkg}=={pkg_version} with CPU succesfully installed!!"
    else:
        error = True
        msg = f"Invalid verion {pkg_version} cuda_version={cuda_version}"

    res = shell(cmd, check=False)
    if res.returncode != 0:
        error = True
        msg = res.stderr.decode()
    return msg, error


if __name__ == "__main__":
    install = os.getenv("INSTALL_TRANSFORMERS")
    if install:
        torch = os.getenv("TORCH_VERSION")
        cuda = os.getenv("CUDA_VERSION")
        res, err = install_pytorch("torch", pkg_version=torch, cuda_version=cuda)
        if err:
            print(res)
            sys.exit(-1)
        print(res)
        sys.exit(0)
    else:
        print("transformers lib not required")



from typing import List, Optional

from pydantic import BaseModel, Field


class GPUInfo(BaseModel):
    device_count: int = 0
    devices: List[str] = Field(default_factory=list)


def get_gpu_info(framework="pt") -> Optional[GPUInfo]:
    import torch

    available = torch.cuda.is_available()
    if not available:
        return None
    count = torch.cuda.device_count()
    devices = [torch.cuda.get_device_name(i) for i in range(0, count)]
    return GPUInfo(device_count=count, devices=devices)



import sqlite3
import warnings
from collections import Counter
from typing import Dict, List

import numpy as np
from pydantic import BaseModel
from sklearn.metrics.pairwise import cosine_similarity

from datawords import constants, parsers


class Entity(BaseModel):
    txt: str
    start_char: int
    end_char: int
    label: str


def extract_entities(doc) -> List[Entity]:
    """Gets a spacy doc object"""
    _entities = []
    for ent in doc.ents:
        e = Entity(ent.text, ent.start_char, ent.end_char, ent.label_)
        _entities.append(e)
    return _entities


def entities_by_label(entities: List[Entity]) -> Dict[str, List[str]]:
    by_label = {}
    for e in entities:
        if by_label.get(e.label):
            by_label[e.label].append(e.txt)
        else:
            by_label.update({e.label: [e.txt]})

    return by_label


def _run_query(cursor: sqlite3.Cursor, q: str):
    """sqlite helper to run queries."""
    return cursor.execute(q).fetchall()


def _longest(data):
    """
    Helper to choose the string with more words:
    [('Maria Perez',), ('Maria Perez de Gozales',), ('Maria',)]
    it will return the second one.
    """
    aux = 0
    final = None
    for x in data:
        if len(x[0].split()) > aux:
            final = x
            aux = len(x)
    return final


def compute_similarity_m(vects):
    """
    It gets a cosine similarity matrix
    the diagonal where each element match with him self is
    filled with 0 values. if values are not zeros
    then pagerank will fail
    """
    A = np.array(vects)
    # A_sparse = sparse.csr_matrix(A)
    M = cosine_similarity(A)

    for i in range(len(vects)):
        for j in range(len(vects)):
            if i == j:
                M[i][j] = 0.0
    return M


def norm_l2_np(vec):
    """
    Normalize a vector. Similar to pytorch norm.
    """
    return vec / np.linalg.norm(vec)


def norm_l2_np_with_zeros(vec):
    """
    If a vector is all zeros it will avoid the normalization
    """
    if not vec.sum() == 0.0:
        return vec / np.linalg.norm(vec)
    return vec


class WordSearch:
    def __init__(self, sqlite=":memory:", stopwords=set()):
        """
        Based  on a list we want to get uniques names ready to be used
        with fts5 type data from sqlite3 which allow us to perform fuzzy
        search over a set of data.
        """
        self.db = sqlite3.connect(sqlite)
        self._create_tables()
        self._stopw = stopwords

    def _parse(self, txt) -> List[str]:
        tokens = parsers.doc_parser(
            txt,
            self._stopw,
            emo_codes=False,
            strip_accents=True,
            lower=True,
            numbers=True,
            parse_urls=False,
        )
        return tokens

    @staticmethod
    def _run_query(cursor: sqlite3.Cursor, q: str):
        """sqlite helper to run queries."""
        return cursor.execute(q).fetchall()

    def _create_tables(self):
        cur = self.db.cursor()
        cur.execute('create virtual table vtags using fts5(name, tokenize="ascii");')
        cur.execute(
            'create virtual table unique_words using fts5(name, tokenize="ascii");'
        )
        cur.execute(
            """CREATE TABLE normal
        (id INTEGER PRIMARY KEY,name TEXT NOT NULL UNIQUE);"""
        )
        cur.close()

    def list_words(self, table="vtags") -> List[str]:
        cur = self.db.cursor()
        rows = cur.execute(f"select * from {table};")

        return [r[0] for r in rows]

    def add(self, word: str) -> bool:
        cur = self.db.cursor()
        added = False
        try:
            cur.execute("insert into normal (name) values (?);", (word,))
            cur.execute("insert into vtags (name) values (?);", (word,))
            self.db.commit()
            added = True
        except sqlite3.IntegrityError:
            pass
        except sqlite3.OperationalError:
            pass

        cur.close()
        return added

    def add_batch(self, words: List[str]) -> List[bool]:
        cur = self.db.cursor()
        tracking = []
        for w in words:
            try:
                cur.execute("insert into normal (name) values (?);", (w,))
                cur.execute("insert into vtags(name) values(?);", (w,))
                tracking.append(True)
            except sqlite3.IntegrityError:
                tracking.append(False)
        self.db.commit()
        cur.close()
        return tracking

    def get(self, word: str, limit: int = 1) -> List[str]:
        warnings.warn("WordsSearch.get() will be deprecated", DeprecationWarning)
        return self.fuzzy_search(word, limit)

    def fuzzy_search(self, name: str, limit: int = 1) -> List[str]:
        cur = self.db.cursor()
        try:
            result = self._run_query(
                cur,
                f"""select * from vtags where name MATCH '"{name}" *'
                                    limit {limit}""",
            )
        except sqlite3.OperationalError:
            result = []
        cur.close()
        return [r for r, in result]

    def ngram_search(self, txt: str, ngrams=2, table="vtags") -> Counter:
        """
        It's an approximate counter of repeated words in a text.

        It counts how often each word in the corpus appears in a table.
        Based on the names previously loaded with `get_uniques`
        this method receive a text and counts how often each name appears
        """
        cur = self.db.cursor()
        c = Counter()
        tokens = self._parse(txt)
        ngrams = parsers.generate_ngrams(tokens, n=ngrams)
        for x in ngrams:
            result = _run_query(
                cur,
                f"""select * from {table} where name MATCH '"{x}" *'
                        limit 1""",
            )
            if len(result) > 0:
                c.update(result[0])

        return c




import json
from typing import Any, Callable, Dict, Optional

from annoy import AnnoyIndex
from pydantic import BaseModel

from datawords.models_old import WordActor, create_word_actor


class IndexConfig(BaseModel):

    words_model: str
    phrases_model: Optional[str] = None
    lang: str = "en"
    vector_size: int = 100
    ann_trees: int = 10
    distance_metric: str = "angular"


class Index:

    def __init__(self, ix: AnnoyIndex,
                 *,
                 words_model: WordActor,
                 id_mapper: Dict[int, Any],
                 conf: IndexConfig,
                 ):
        self.ix = ix
        self.id_mapper = id_mapper
        self.words = words_model
        self.conf = conf

    def encode(self, txt):
        return self.words.encode(txt)

    def save(self, fpath: str):
        self.ix.save(f"{fpath}.ann")
        with open(f"{fpath}.conf.json", "w") as f:
            f.write(self.conf.json())

        with open(f"{fpath}.maps.json", "w") as f:
            f.write(json.dumps(self.id_mapper))

    def search(self, txt, top_n=5, include_distances=False):
        v = self.encode(txt)
        vectors = self.ix.get_nns_by_vector(
            v, top_n, include_distances=include_distances)
        if include_distances:
            _res = [self.id_mapper[_v] for _v in vectors[0]]
            res = list(zip(_res, vectors[1]))
        else:
            res = [self.id_mapper[_v] for _v in vectors]
        return res


def load(bpath: str, *, index_name,
         words_model: Optional[WordActor] = None) -> Index:
    with open(f"{bpath}/models/{index_name}.conf.json", "r") as f:
        data = f.read()
        jdata = json.loads(data)
        conf = IndexConfig(**jdata)
    if not words_model:
        words_model = create_word_actor(
            bpath,
            wv_model=conf.words_model,
            phrases_model=conf.phrases_model,
            lang=conf.lang,
        )
    with open(f"{bpath}/models/{index_name}.maps.json", "r") as f:
        _ids = json.loads(f.read())
        ids_mapped = {int(k): v for k, v in _ids.items()}

    t = AnnoyIndex(conf.vector_size, conf.distance_metric)
    t.load(f"{bpath}/models/{index_name}.ann")
    ix = Index(
        t, words_model=words_model,
        id_mapper=ids_mapped,
        conf=conf,
    )
    return ix


def build_index(
        base_path: str,
        *,
        words_model: Optional[WordActor] = None,
        words_model_name: str,
        words_phrases: str = None,
        lang: str = "en",
        data_ids,
        data_getter: Callable,
        vector_size: int = 100,
        ann_trees=10,
        distance="angular"
) -> Index:
    if not words_model:
        words_model = create_word_actor(base_path,
                                        wv_model=words_model_name,
                                        phrases_model=words_phrases,
                                        lang=lang
                                        )
    f = vector_size
    t = AnnoyIndex(f, distance)
    id_mapper = {}
    for _ix, _id in enumerate(data_ids):
        data = data_getter(_id)
        v = words_model.encode(data)
        t.add_item(_ix, v)
        id_mapper[_ix] = _id

    t.build(ann_trees)
    ix = Index(t, words_model=words_model,
               id_mapper=id_mapper,
               conf=IndexConfig(
                   words_model=words_model_name,
                   phrases_model=words_phrases,
                   lang=lang,
                   vector_size=vector_size, ann_trees=ann_trees,
                   distance_metric=distance
               )
               )
    return ix



import json
import os
from typing import Iterable, List, Optional, Union

import numpy as np
from gensim.models import KeyedVectors, Word2Vec
from gensim.models.phrases import FrozenPhrases, Phrases
from pydantic import BaseModel

from datawords import _utils, constants, parsers


class PhrasesModelMeta(BaseModel):
    name: str
    lang: str
    parser_conf: parsers.ParserConf
    min_count: float = 1.0
    threshold: Optional[float] = None
    max_vocab_size: Optional[int] = None
    version: str = _utils.get_version()


class W2VecMeta(BaseModel):
    name: str
    lang: str
    parser_conf: parsers.ParserConf
    phrases_model_path: Optional[str] = None
    epoch: int = 5
    size: int = 100
    window: int = 5
    min_count: int = 1
    version: str = _utils.get_version()


class PhrasesModel:
    def __init__(
        self,
        parser_conf: parsers.ParserConf,
        min_count: float = 1.0,
        threshold: Optional[float] = None,
        max_vocab_size: Optional[int] = None,
        connector_words=None,
        model=None,
    ):
        """

        :param parser_conf: configuration of the parser
        :type parser_conf: parsers.ParserConf
        :param min_count: Ignore all words and bigrams with total collected count
            lower than this value.
        :type min_count: float
        :param threshold: Represent a score threshold for forming the phrases
            (higher means fewer phrases). A phrase of words a followed by b is
            accepted if the score of the phrase is greater than threshold. Heavily
            depends on concrete scoring-function, see the scoring parameter.
        :param max_vocab_size: Maximum size (number of tokens) of the vocabulary.
            Used to control pruning of less common words, to keep memory under
            control. The default of 40M needs about 3.6GB of RAM.
            Increase/decrease max_vocab_size depending on how much available
            memory you have.
        :type max_vocab_size: Optional[int] int
        :param connector_words: Set of words that may be included within a phrase,
             without affecting its scoring. If any is provided it will use the
             lang value from the parser_conf. By default datawords
             include CONNECTOR_WORDS for English, Portugues an Spanish.
        :type connector_words: Frozenset[str]

        """

        self._parser_conf = parser_conf
        self._min_count = min_count
        self._threshold = threshold
        self._max_vocab_size = max_vocab_size
        self._connector_words = (
            connector_words or constants.CONNECTOR_WORDS[parser_conf.lang]
        )
        self._model = model
        self._parser = self._parser_factory()

    @property
    def model(self) -> FrozenPhrases:
        return self._model

    def _load_stopw(self):
        return parsers.load_stop2(
            self._parser_conf.lang, models_path=self._parser_conf.stopw_path
        )

    def _parser_factory(self) -> parsers.SentencesParser:
        stopw = self._load_stopw()
        return parsers.SentencesParser(
            lang=self._parser_conf.lang,
            lower=self._parser_conf.lower,
            emo_codes=self._parser_conf.emo_codes,
            strip_accents=self._parser_conf.strip_accents,
            numbers=self._parser_conf.numbers,
            stop_words=stopw,
        )

    def fit(self, X: Iterable):
        """
        This will train the phrase model. It needs an iterable.

        :param X: An iterable which returns plain texts.
        :type X: Iterable
        """
        sentences = parsers.SentencesIterator(X, parser=self._parser)
        _model = Phrases(
            sentences,
            min_count=self._min_count,
            threshold=1,
            connector_words=self._connector_words,
        )
        self._model = _model.freeze()

    def transform(self, X: Iterable) -> List[List[str]]:
        """
        Trasform a list of texts.

        :param X: an iterable.
        :type X: Iterable

        :return: A list of phrases.
        :rtype: List[List[str]]
        """
        results = []
        for txt in X:
            r = self._model[txt.split()]
            results.append(r)
        return results

    def parse(self, txt: str) -> List[str]:
        """
        It will parse only one text.
        :param txt: str
        :return: a list of words
        :rtype: List[str]
        """
        r = self._model[txt.split()]
        return r

    def save(self, fp: Union[str, os.PathLike]):
        """
        Save phrase model to a folder.

        :param fp: The path to the folder. Each model is stored in a folder.
            The path should be to that folder.
        :type fp: Union[str, os.PathLike]
        """

        name = str(fp).rsplit("/", maxsplit=1)[1]
        _utils.mkdir_p(fp)

        conf = PhrasesModelMeta(
            name=name,
            lang=self._parser_conf.lang,
            parser_conf=self._parser_conf,
            min_count=self._min_count,
            threshold=self._threshold,
        )
        self._model.save(f"{fp}/{name}.bin")
        with open(f"{fp}/{name}.json", "w") as f:
            f.write(conf.json())

    @classmethod
    def load(cls, fp: Union[str, os.PathLike]) -> "PhrasesModel":
        """loads the TextIndex model.

        :param fp: The path to the index. Each model is stored in a folder.
             The path should be to that folder.
        :type fp: Union[str, os.PathLike]
        :return: PhrasesModel loaded.
        :rtype: PhrasesModel
        """

        name = str(fp).rsplit("/", maxsplit=1)[1]
        with open(f"{fp}/{name}.json", "r") as f:
            jmeta = json.loads(f.read())
            meta = PhrasesModelMeta(**jmeta)
        model = Phrases.load(f"{fp}/{name}.bin")
        obj = cls(
            parser_conf=meta.parser_conf,
            min_count=meta.min_count,
            threshold=meta.threshold,
            model=model,
        )
        return obj


class Word2VecHelper:
    def __init__(
        self,
        parser_conf: parsers.ParserConf,
        phrases_model=None,
        size: int = 100,
        window: int = 5,
        min_count: int = 1,
        workers: int = 1,
        epoch: int = 5,
        model: Word2Vec = None,
        using_kv=False,
        loaded_from=None,
    ):
        """

        :param parser_conf: configuration of the parser
        :type parser_conf: parsers.ParserConf
        :param min_count: Ignore all words and bigrams with total collected
            count lower than this value.
        :type min_count: Optional[float]
        :param threshold: Represent a score threshold for forming the phrases
            (higher means fewer phrases). A phrase of words a followed by b is
            accepted if the score of the phrase is greater than threshold.
            Heavily depends on concrete scoring-function, see the scoring
            parameter.
        :param max_vocab_size: Maximum size (number of tokens) of the
            vocabulary. Used to control pruning of less common words,
            to keep memory under control. The default of 40M needs
            about 3.6GB of RAM. Increase/decrease max_vocab_size
            depending on how much available memory you have.
        :type max_vocab_size: Optional[int] int
        :param connector_words: Set of words that may be included within
           a phrase, without affecting its scoring. If any is provided it
           will use the lang value from the parser_conf. By default
           datawords include CONNECTOR_WORDS for English, Portugues
           an Spanish.
        :type connector_words: Frozenset[str]

        """

        self._parser_conf = parser_conf
        self._stopw = self._load_stopw(parser_conf.lang, self._parser_conf.stopw_path)
        self._phrases = phrases_model
        # self._models_path = models_path
        self._min_count = min_count
        self._size = size
        self._window = window
        self._min_count = min_count
        self._workers = workers
        self._epoch = epoch
        self._parser = self._parser_factory()
        self.model: Union[Word2Vec, KeyedVectors] = model
        self._using_kv = using_kv
        self.loaded_from = loaded_from

    @property
    def vector_size(self) -> int:
        return self._size

    @property
    def wv(self) -> Union[Word2Vec, KeyedVectors]:
        return self.model

    def _parser_factory(self) -> parsers.SentencesParser:
        return parsers.SentencesParser(
            lang=self._parser_conf.lang,
            lower=self._parser_conf.lower,
            phrases_model=self._phrases,
            emo_codes=self._parser_conf.emo_codes,
            strip_accents=self._parser_conf.strip_accents,
            numbers=self._parser_conf.numbers,
            stop_words=self._stopw,
        )

    def _load_stopw(self, lang, models_path):
        return parsers.load_stop2(lang, models_path=models_path)

    def fit(self, X: Iterable):
        """
        This will train the model. It needs an iterable.

        :param X: An iterable which returns plain texts.
        :type X: Iterable
        """

        sentences = parsers.SentencesIterator(X, parser=self._parser)
        model = Word2Vec(
            sentences=sentences,
            vector_size=self._size,
            window=self._window,
            min_count=self._min_count,
            workers=self._workers,
        )
        self.model = model

    # def transform(self, X: Iterable) -> np.ndarray:
    #     """
    #     This will train the model. It needs an iterable.

    #     :param X: An iterable which returns plain texts.
    #     :type X: Iterable
    #     """

    #     sentences = parsers.SentencesIterator(X, parser=self._parser)
    #     vectors = [self.encode(st) for st in sentences]
    #     return np.asarray(vectors)

    # def fit_transform(self, X: Iterable) -> np.ndarray:
    #     self.fit(X)
    #     return self.transform(X)

    def parse(self, sentence: str) -> List[str]:
        """
        It will parse only one text.
        :param txt: str
        :return: a list of words
        :rtype: List[str]
        """
        return self._parser.parse(sentence)

    def encode(self, sentence: str) -> np.ndarray:
        """
        gets a sentence in plain text and encode it as vector
        """
        words = self.parse(sentence)
        v = self.vectorize(words)
        return v

    def vectorize(self, sentence: List[str]) -> np.ndarray:
        """Get a vector from a list of words
        if a sentence has words that don't match in the word2vec model,
        then it fills with zeros
        """
        vectors = []
        for word in sentence:
            try:
                _vect = self.model.wv[word]
                vectors.append(_vect)
            except KeyError:
                pass
        if len(vectors) > 0:
            return np.sum(np.array(vectors), axis=0) / (len(vectors) + 0.001)
        return np.zeros((self._size,))

    def save(self, fp: Union[str, os.PathLike]):
        name = str(fp).rsplit("/", maxsplit=1)[1]
        _utils.mkdir_p(fp)
        conf = W2VecMeta(
            name=name,
            lang=self._parser_conf.lang,
            parser_conf=self._parser_conf,
            size=self._size,
            window=self._window,
            min_count=self._min_count,
            epoch=self._epoch,
        )

        self.model.save(f"{fp}/{name}.bin")
        self.model.wv.save(f"{fp}/{name}.kv")
        with open(f"{fp}/{name}.json", "w") as f:
            f.write(conf.json())

    @classmethod
    def load(cls, fp: Union[str, os.PathLike], keyed_vectors=False) -> "Word2VecHelper":
        name = str(fp).rsplit("/", maxsplit=1)[1]
        if keyed_vectors:
            model = KeyedVectors.load(f"{fp}/{name}.kv", mmap="r")
        else:
            model = Word2Vec.load(f"{fp}/{name}.bin")
        with open(f"{fp}/{name}.json", "r") as f:
            jmeta = json.loads(f.read())
            meta = W2VecMeta(**jmeta)
        phrases = None
        if meta.phrases_model_path:
            phrases = PhrasesModel.load(meta.phrases_model_path)
        obj = cls(
            parser_conf=meta.parser_conf,
            phrases_model=phrases,
            size=meta.size,
            window=meta.window,
            min_count=meta.min_count,
            model=model,
            using_kv=keyed_vectors,
            loaded_from=str(fp),
        )
        return obj



from typing import List, Optional

import numpy as np
import torch

import torch.nn.functional as F
from transformers import (
    AutoModel,
    AutoTokenizer,
    BertModel,
    BertTokenizerFast,
    pipeline,
)

def is_cuda_available() -> bool:
    return torch.cuda.is_available()


def norm_l2_torch(vect):
    norm = F.normalize(vect, p=2)
    return norm


def to_np_arr(vect, l2_norm=True):
    if l2_norm:
        norm = norm_l2_torch(vect)
        return norm.detach.numpy()
    return vect.detach.numpy()


class Transformer:
    """ A Bag for differents NLP tasks using transformers from
    HuggingFace and Spacy.
    Tasks:
       - Gets embeddings for similarity and clustering tasks.
       - Classify texts based on labels using Zero-shot technique.
       - Entity extraction using spacy library.
    """

    def __init__(self, models_path: Optional[str]=None):
        # classifier and encoder based on NLI Distil Roberta
        self.nli_model = None
        self.nli_tokenizer = None
        self.classifier = None
        self.features_extractor = None

        self.models_path = models_path
        self.nlp = None
        self.nlp_en = None

        self.labse = None
        self.labse_tokenizer = None

    def _get_fullpath(self, model_name: str) -> str:
        fp = model_name
        if self.models_path:
            fp = f"{self.models_path}/{model_name}"
        return fp

    def load_labse(self, model_name: str = "setu4993/smaller-LaBSE"):
        """ labse is a multilengual model that can get similar vector representation for
        differents langs. """
        fullpath = self._get_fullpath(model_name)
        tokenizer = BertTokenizerFast.from_pretrained(fullpath)
        model = BertModel.from_pretrained(fullpath)
        model = model.eval()
        self.labse = model
        self.labse_tokenizer = tokenizer

    def multilingual_vectors(self, sentences, l2_norm=False):
        inputs = self.labse_tokenizer(
            sentences, return_tensors="pt", padding=True)

        with torch.no_grad():
            outputs = self.labse(**inputs)

        embeddings = outputs.pooler_output
        if l2_norm:
            embeddings = F.normalize(embeddings, p=2)
        return embeddings

    def load_nli_model(self,
                       model_name="cross-encoder/nli-distilroberta-base"):
        fullpath = self._get_fullpath(model_name)
        self.nli_tokenizer = AutoTokenizer.from_pretrained(fullpath)
        self.nli_model = AutoModel.from_pretrained(fullpath)

    def load_classifier(self,
                        model_name="cross-encoder/nli-distilroberta-base"):
        fullpath = self._get_fullpath(model_name)

        self.classifier = pipeline("zero-shot-classification", model=fullpath)
        # tokenizer=self.nli_tokenizer)

    def load_features_extractor(self,
                                model_name="cross-encoder/nli-distilroberta-base"):
        if not self.nli_model:
            self.load_nli_model(model_name=model_name)

        self.features_extractor = pipeline("feature-extraction", model=self.nli_model,
                                           tokenizer=self.nli_tokenizer)

    def get_embeddings(self, text):
        rsp = self.features_extractor(text)
        return rsp[0][0]

    def classify(self, texts: List[str], labels: List[str]):
        return self.classifier(texts, labels)





# from scipy import sparse
from typing import Any, List

import numpy as np
from annoy import AnnoyIndex
from scipy import sparse
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sknetwork.ranking import PageRank
from tqdm import tqdm

from datawords.utils import norm_l2_np_with_zeros


def get_adj(edges, size):
    E = edges
    # size = max(max(E))+1
    r = [[0 for i in range(size)] for j in range(size)]
    for row, col in E:
        r[row][col] = 1
    return np.asarray(r)


def build_annoy_edges(ix: AnnoyIndex, vectors: np.ndarray, barrier=0.9, tqdm_=True):
    edges = []
    if tqdm_:
        for i in tqdm(range(vectors.shape[0])):
            for j in range(vectors.shape[0]):
                if i != j:
                    dst = ix.get_distance(i, j)
                    if dst <= barrier:
                        edges.append((i, j))
    else:
        for i in range(vectors.shape[0]):
            for j in range(vectors.shape[0]):
                if i != j:
                    dst = ix.get_distance(i, j)
                    if dst <= barrier:
                        edges.append((i, j))

    return edges


class PageRankTFIDF:
    # pylint: disable=too-few-public-methods
    """PageRank based on TFIDF"""

    def __init__(
        self, tokenizer=None, barrier=0.7, min_df=1, max_df=0.9, ngram_range=(1, 1)
    ):
        # pylint: disable=too-many-arguments
        self.tf = TfidfVectorizer(
            analyzer="word",
            ngram_range=ngram_range,
            # stop_words=stopw,
            tokenizer=tokenizer,
            min_df=min_df,
            max_df=max_df,
        )
        self.barrier = barrier

    @staticmethod
    def _compute_tfidf_M(A):
        M = cosine_similarity(A)

        for i in range(A.shape[0]):
            for j in range(A.shape[0]):
                if i == j:
                    M[i][j] = 0.0
        return M

    @staticmethod
    def _build_edges_all(A, barrier=0.7):
        edges = []
        M = cosine_similarity(A)

        for i in range(A.shape[0]):
            for j in range(A.shape[0]):
                if i != j:
                    cosine = M[i, j]
                    if cosine > barrier:
                        # edges.append((i, j, cosine))
                        edges.append((i, j))

        return edges

    def fit_transform(self, corpus):
        """
        Estimate a pagerank by cosine_similarity
        """
        tf_M = self.tf.fit_transform(corpus)
        sim_M = self._compute_tfidf_M(tf_M)
        adjacency = sparse.csr_matrix(sim_M)
        pagerank = PageRank()
        scores = pagerank.fit_transform(adjacency)
        return scores


# corpus: Union[Generator, List[str]]
class PageRankAnnoy:
    # pylint: disable=too-many-instance-attributes
    """
    It uses Annoy Indexer and PageRank to rank a list of `documents`,
    where each document should be an string.

    PageRank is an old known google algorithm, originally used to get the
    "most important" links from internet.

    In the text world, there are different approachs to achive the same goal.
    For instance, if two documents share the same words, then they could
    be considered as nodes connected between.

    Because PageRank is a graph algorithm the idea of nodes
    and edges is important.

    Other approach is to measure the similarity between
    two documents (nodes), if they are similar enough (`barrier`),
    then they are connected. This last method is used here.

    """

    def __init__(
        self,
        metric_distance="euclidean",
        l2_norm=True,
        barrier=0.9,
        n_trees=10,
        n_jobs=-1,
        tqdm=True,
    ):
        """
        For `l2_norm` param refer to
        https://machinelearningmastery.com/vector-norms-machine-learning/

        :param metric_distance: the options are  "angular", "euclidean",
            "manhattan", "hamming", or "dot". It will be used by Annoy
             as measure to calculates distance between texts.

        :param l2_norm: True by default, usually is recommend their usage.
        :param barrier: It depends on the metric_distance choose, but
            this param will serve as filter to define if 2 texts are
            connected as nodes.
        :param n_trees: trees used by annoy index.
        :param n_jobs: using multithreading for the index.
        :param tqdm: tqdm usage when adj. matrix is calculated.

        """
        self.barrier = barrier
        self.metric_distance = metric_distance
        self.n_trees = n_trees
        self.n_jobs = n_jobs
        self.X = None
        self.aix = None
        self._adj = None
        self._edges = None
        self._scores = None
        self._tqdm = tqdm
        self._l2_norm = l2_norm
        # self.annoy_ix = AnnoyIndex(vector_size, metric_distance)
        # self.vectors = None

    @staticmethod
    def as_ndarray(X: List[np.ndarray]) -> np.ndarray:
        return np.asarray(X)

    def fit(self, X: np.ndarray):
        self.X = X
        if self._l2_norm:
            self.X = np.apply_along_axis(norm_l2_np_with_zeros, 1, X)

    def transform(self):
        self.aix = AnnoyIndex(self.X.shape[1], self.metric_distance)
        for i in range(self.X.shape[0]):
            self.aix.add_item(i, self.X[i])

        self.aix.build(n_trees=self.n_trees, n_jobs=self.n_jobs)
        pg = PageRank()
        self._edges = build_annoy_edges(
            self.aix, self.X, barrier=self.barrier, tqdm_=self._tqdm
        )
        self._adj = get_adj(self._edges, self.X.shape[0])

        self._scores = pg.fit_transform(self._adj)
        return self._scores

    def fit_transform(self, X: np.ndarray):
        self.fit(X)
        scores = self.transform()
        return scores

    def rank(self, index: List[Any], top_n=5):
        """given a `index` build a ranking using the scores created
        by PageRank."""
        ranking = zip(index, self.scores)
        best = sorted(ranking, key=lambda tup: tup[1], reverse=True)[:top_n]
        return best

    @property
    def scores(self):
        return self._scores

    @property
    def edges(self):
        return self._edges

    @property
    def adjacency(self):
        return self._adj


def rank(index: List[Any], scores: np.ndarray, top_n=5):
    """
    given a score result from any of the PageRank models
    an a index which correlates with the data used to calculate
    pagerank then, reorder score from top to bottom.

    To produce a simple numeric index, :function:`datawords.ranking.numeric_index`
    could be used. 
    """
    ranking = zip(index, scores)
    best = sorted(ranking, key=lambda tup: tup[1], reverse=True)
    if top_n:
        return best[:top_n]
    return best


def numeric_index(values: List[Any]) -> List[int]:
    """
    given a list of values of any type it produce
    an incremental index of int.
    useful for :function:`datawords.ranking.rank`
    """
    return list(range(0, len(values)))



import numpy as np

from datawords import models, parsers, ranking
from .shared import open_texts


def test_ranking_pagerank_annoy():
    texts = list(open_texts())
    parser_conf = parsers.ParserConf(lang="en")
    wv = models.Word2VecHelper(parser_conf)
    wv.fit(texts)
    vects = [wv.encode(t) for t in texts]

    pra = ranking.PageRankAnnoy()
    arrs = pra.as_ndarray(vects)
    scores = pra.fit_transform(arrs)
    index = [ix for ix, _ in enumerate(texts)]
    rank = pra.rank(index)
    assert len(rank) == 5
    assert isinstance(scores, np.ndarray)


def test_ranking_tfidf():
    texts = list(open_texts())
    pr = ranking.PageRankTFIDF()
    scores = pr.fit_transform(texts)
    index = ranking.numeric_index(texts)
    rank = ranking.rank(index, scores)
    assert len(rank) == 5
    assert isinstance(scores, np.ndarray)



import os
from pathlib import Path


def pkg_route() -> str:
    """
    Get the absoute path of the labfunctions package whatever it is installed
    It could be used to reference files inside of the package.
    :return:
    :type str:
    """
    here = os.path.abspath(os.path.dirname(__file__))
    return here


def get_version() -> str:
    # fp = pkg_route()
    # with open(f"{fp}/__about__.py", "r") as f:
    #     f.read()
    from datawords.__about__ import __version__

    return __version__


def mkdir_p(fp):
    """Make the fullpath
    similar to mkdir -p in unix systems.
    """
    Path(fp).mkdir(parents=True, exist_ok=True)



from typing import List

from toolz import partition_all
from tqdm.auto import tqdm
from transformers import MarianMTModel, MarianTokenizer, pipeline

from datawords.constants import ROMANCE_SUPPORT
from datawords.deepnlp.utils import get_gpu_info


class TranslatorBeta:
    ROMANCE_SUPPORT = ROMANCE_SUPPORT

    def __init__(
        self,
        source,
        target,
        *,
        model_path=None,
        limit_text=250,
        max_length=512,
        device=-1,
        automatic_gpu_detection=True,
        progress_bar=True,
        chunk_size=50,
    ):
        """
        :param source: lang source, it could be any ROMANCE lang, or `en`.
        :type source: str
        :param target: Any ROMANCE or en lang.
        :type target: str

        """

        if model_path:
            fullpath = model_path
        else:
            fullpath = self.build_model_name(source, target)
        self._source = source
        self._target = target
        self._limit = limit_text
        self.tokenizer = MarianTokenizer.from_pretrained(
            fullpath, model_max_length=max_length, truncation=True
        )
        self.model = MarianMTModel.from_pretrained(fullpath)
        _device = self._get_device() if automatic_gpu_detection else device
        self.using_device = _device
        self.pipe = pipeline(
            f"translation_{source}_to_{target}",
            model=self.model,
            tokenizer=self.tokenizer,
            device=_device,
        )
        self._progress_bar = progress_bar
        self._chunk = chunk_size

    def _get_device(self) -> int:
        gpu = get_gpu_info()
        if gpu:
            return 0
        return -1

    def transform(self, texts: List[str]) -> List[str]:
        """
        TODO: check how to bring a generator.
        Using pytorch dataloaders

        or as a quick hack gets an iterator, and then
        implement pipe in chunks
        """
        _disable = not self._progress_bar
        results = []
        chunks = partition_all(self._chunk, texts)
        total = len(texts) / self._chunk
        for chunk in tqdm(chunks, disable=_disable, total=total):
            for translated in self.pipe(list(chunk)):
                results.append(translated["translation_text"])
        return results

    @staticmethod
    def build_model_name(source, target) -> str:
        _src = source
        _target = target
        if source in ROMANCE_SUPPORT:
            _src = "ROMANCE"
        elif target in ROMANCE_SUPPORT:
            _target = "ROMANCE"
        else:
            raise NameError(f"Model not found for {source} nor {target}")
        # return f"{base_path}/{model_name}"
        return f"Helsinki-NLP/opus-mt-{_src}-{_target}"



from typing import List

import numpy as np
from gensim.models import KeyedVectors, Word2Vec
from gensim.models.phrases import FrozenPhrases, Phrases
from sklearn.metrics.pairwise import cosine_similarity

# import spacy
from datawords.parsers import doc_parser, load_stop

# from sknetwork.ranking import PageRank


def compute_similarity_m(vects):
    """get a cosine similarity matrix
    the diagonal where each element match with their self is
    matched to 0, if not pagerank fail
    """
    A = np.array(vects)
    # A_sparse = sparse.csr_matrix(A)
    M = cosine_similarity(A)

    for i in range(len(vects)):
        for j in range(len(vects)):
            if i == j:
                M[i][j] = 0.0
    return M


class WordActor:

    def __init__(self, stopw, wv_model: Word2Vec, nlp=None,
                 phrases: FrozenPhrases = None):
        self.nlp = nlp
        self.stopw = stopw
        self.wv: Word2Vec = wv_model
        self.vector_size = self.wv.vector_size
        self.phrases = phrases

    def load_nlp(self, nlp):
        self.nlp = nlp

    def doc_parser(self, txt, strip_accents=True):
        return doc_parser(txt, self.stopw, strip_accents=strip_accents)

    def get_vector2(self, s: List[str]):
        """ Get a vector from a list of words
        if a sentence doesn't match, then it fills with zeros
        """
        vectors = []
        for word in s:
            try:
                _vect = self.wv[word]
                vectors.append(_vect)
            except KeyError:
                pass
        if len(vectors) > 0:
            return np.sum(np.array(vectors), axis=0) / (len(vectors) + 0.001)
        else:
            return np.zeros((self.vector_size,))

    def encode(self, txt: str):
        """
        gets a sentence in plain text and encode it as vector
        """
        words = self.doc_parser(txt)
        if self.phrases:
            sentence = self.phrases[words]
        else:
            sentence = words
        v = self.get_vector2(sentence)
        return v

    # def pagerank(self, df, column="title", strip_accents=True):
    #     """
    #     Estimate a pagerank by cosine_similarity
    #     """
    #     vects, _ = self.sentences(df, column, strip_accents)
    #     sim_M = compute_similarity_m(vects)
    #     adjacency = sparse.csr_matrix(sim_M)
    #     pagerank = PageRank()
    #     scores = pagerank.fit_transform(adjacency)
    #     return vects, scores

    def init_nlp_doc(self, text: str):
        return self.nlp(text)

    def similarity(self, txt1, txt2, strip_accents=True):
        vec1 = self.get_vector2(doc_parser(txt1,
                                           self.stopw,
                                           strip_accents=strip_accents))
        vec2 = self.get_vector2(doc_parser(txt2, self.stopw,
                                           strip_accents=strip_accents))
        simil = cosine_similarity([vec1], [vec2])
        return simil[0][0]


def create_word_actor(base_path: str,
                      wv_model: str,
                      phrases_model: str = None,
                      lang="en",
                      nlp_model_name=None,
                      nlp_rank=False) \
        -> WordActor:
    """ Helper function for a singleton word actor model """
    # pylint: disable=maybe-no-member
    # pylint: disable=
    nlp = None
    stopw = load_stop(
        f"{base_path}/models", lang=lang)

    wv_path = f"{base_path}/models/{wv_model}"
    wv = KeyedVectors.load(wv_path, mmap='r')

    if nlp_model_name:
        import spacy
        nlp = spacy.load(f"{base_path}/models/{nlp_model_name}")
        if nlp_rank:
            import pytextrank
            nlp.add_pipe("textrank")

    phrases = None
    if phrases_model:
        phrases = Phrases.load(f"{base_path}/models/{phrases_model}")
    actor = WordActor(stopw, wv_model=wv.wv, nlp=nlp, phrases=phrases)

    return actor



import tempfile

from gensim.models import KeyedVectors, Word2Vec
from gensim.models.phrases import FrozenPhrases

from datawords import parsers
from datawords.models import PhrasesModel, Word2VecHelper


def open_texts():
    with open("tests/texts.txt", "r") as f:
        texts = f.readlines()
    for t in texts:
        _t = t.strip()
        if _t:
            yield _t



def test_models2_phrases():
    texts = open_texts()
    parser_conf = parsers.ParserConf(lang="en")
    model = PhrasesModel(parser_conf)
    model.fit(texts)

    with tempfile.TemporaryDirectory() as tmpdir:
        fp = f"{tmpdir}/tests"
        model.save(fp)
        model2 = PhrasesModel.load(fp)

    assert isinstance(model.model, FrozenPhrases)
    assert isinstance(model2.model, FrozenPhrases)


def test_models2_word2vec():
    texts = open_texts()
    parser_conf = parsers.ParserConf(lang="en")
    # model = PhrasesModel(parser_conf)
    # model.fit(texts)
    wv = Word2VecHelper(parser_conf)
    wv.fit(texts)
    with tempfile.TemporaryDirectory() as tmpdir:
        fp = f"{tmpdir}/tests"
        wv.save(fp)
        wv2 = Word2VecHelper.load(fp)
        kv = Word2VecHelper.load(fp, keyed_vectors=True)

    assert isinstance(wv.model, Word2Vec)
    assert isinstance(wv2.model, Word2Vec)
    assert isinstance(kv.model, KeyedVectors)
